unit dtoolsu;

// Copyright  2000 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

interface

uses Windows, SysUtils, Graphics, Classes, Dialogs, imFrameU;

type
TDrawingTool = class(TObject)
private
  fFore, fBack     : Integer;  //foreground and background colors
  fStartX, fStartY : Integer;  //initial coordinates
  fPrevX, fPrevY   : Integer;  //most recent move-coordinates
  fFrame           : TImageFrame; // image-frame being drawn
  fRBtn            : Boolean;  //true if right button in use
  fDirty           : TNotifyEvent; //Call when a change is made
  procedure DrawLine(x1, y1, x2, y2 : integer; colorIdx : Integer);
  procedure FrameOK;
  procedure XorLines(OnOff : Boolean);
public
  constructor Create(vDirty : TNotifyEvent);
  function  CoordString(X, Y : Integer) : String; virtual;
  procedure DoneUsing; virtual;
  procedure DrawStart(X, Y, fore, back : Integer; RBtn : Boolean;
    vFrame : TImageFrame); virtual;
  procedure DrawEnd(X, Y : Integer); virtual;
  procedure DrawMove(X, Y : Integer; constrain : Boolean); virtual;
  procedure DrawCancel; virtual;
end;

TDPencil = class(TDrawingTool)
  procedure DrawStart(X, Y, fore, back : Integer; RBtn : Boolean;
    vFrame : TImageFrame); override;
  procedure DrawMove(X, Y : Integer; constrain : Boolean); override;
  procedure DrawEnd(X, Y : Integer); override;
end;

TDDropper = class(TDrawingTool)
  procedure DrawEnd(X, Y : Integer); override;
end;

TDPaintCan = class(TDrawingTool)
  procedure DrawEnd(X, Y : Integer); override;
end;

TDrawingToolC = class(TDrawingTool)
//Drawing tool that reports (x,y) -> (x,y) coordinates
private
  procedure SnapRect(X1, Y1: Integer; var X2, Y2: Integer);
public
  function CoordString(X, Y : Integer) : String; override;
end;

TDLine = class(TDrawingToolC)
private
  procedure SnapLine(X1, Y1: Integer; var X2, Y2: Integer);
public
  procedure DrawMove(X, Y : Integer; constrain : Boolean); override;
  procedure DrawEnd(X, Y : Integer); override;
  procedure DrawCancel; override;
end;

TDRect = class(TDrawingToolC)
//Used for rectangle, filled-rect, and chooser
private
  fFill : Boolean;
  function  RectAlign(fL, fT, fR, fB: Integer): TRect;
public
  constructor Create(vDirty : TNotifyEvent);
  procedure DrawMove(X, Y : Integer; constrain : Boolean); override;
  procedure DrawEnd(X, Y : Integer); override;
  procedure DrawCancel; override;
end;

TDRectF = class(TDRect)
  constructor Create(vDirty : TNotifyEvent);
end;

TDChooser = class(TDRect)
private
  fOrgRect : TRect;
public
  procedure DrawStart(X, Y, fore, back : Integer; RBtn : Boolean;
    vFrame : TImageFrame); override;
  procedure DrawEnd(X, Y : Integer); override;
end;

TDEllipse = class(TDrawingToolC)
private
  fFill : Boolean;
public
  constructor Create(vDirty : TNotifyEvent);
  procedure DrawMove(X, Y : Integer; constrain : Boolean); override;
  procedure DrawEnd(X, Y : Integer); override;
  procedure DrawCancel; override;
end;

TDEllipseF = class(TDEllipse)
  constructor Create(vDirty : TNotifyEvent);
end;

TDText = class(TDrawingTool)
public
  procedure DrawEnd(X, Y : Integer); override;
end;

TDMover = class(TDrawingTool)
private
  fValid : Boolean;
  fOrgRect : TRect;
public
  procedure DrawStart(X, Y, fore, back : Integer; RBtn : Boolean;
    vFrame : TImageFrame); override;
  procedure DrawMove(X, Y : Integer; constrain : Boolean); override;
  procedure DrawEnd(X, Y : Integer); override;
  procedure DoneUsing; override;
end;

implementation
uses textfm;

{ TDrawingTool }

constructor TDrawingTool.Create(vDirty : TNotifyEvent);
begin
  Inherited Create;
  fFore  := 0; fStartX := 0; fPrevX  := MaxInt;
  fBack  := 0; fStartY := 0; fPrevY  := MaxInt;
  fFrame := nil;
  fRBtn  := False;
  fDirty := vDirty;
end;

function TDrawingTool.CoordString(X, Y: Integer): String;
begin
  Result := Format('(%d,%d)', [X,Y]);
end;

procedure TDrawingTool.DrawStart(X, Y, fore, back: Integer;
  RBtn : Boolean; vFrame : TImageFrame);
begin
  fStartX := X; fPrevX  := MaxInt;
  fStartY := Y; fPrevY  := MaxInt;
  fRBtn   := RBtn;
  IF fRBtn THEN
    begin
      fFore := Back;
      fBack := Fore;
    end
  ELSE
    begin
      fFore := Fore;
      fBack := Back;
    end;
  fFrame := vFrame;
  FrameOK;
end;

//do-nothing methods in this base object
procedure TDrawingTool.DrawEnd(X, Y: Integer); begin end;
procedure TDrawingTool.DrawMove(X, Y: Integer; constrain: Boolean);
begin end;
procedure TDrawingTool.DrawCancel; begin end;
procedure TDrawingTool.DoneUsing; begin end;

procedure TDrawingTool.DrawLine(x1, y1, x2, y2 : integer;
  colorIdx : Integer);
// A modified version of the well-known Bresenham line-drawing
// algorithm. This version speeds the process by drawing from
// both ends of the line to the middle.
VAR
  Rise     : Integer; // change in Y direction
  Rrun     : Integer; // change in X direction
  LineLen  : Integer; // length in longer direction
  decision : Integer; // : these three decide when to increment
  dIncNeg  : Integer; // : the dependent variable
  dIncPos  : Integer; // :
  xIncNeg  : Integer; // X increment when decision is negative
  yIncNeg  : Integer; // Y increment when decision is negative
  xIncPos  : Integer; // X increment when decision is non-negative
  yIncPos  : Integer; // Y increment when decision is non-negative
  X, Y, N  : Integer;
begin
  FrameOK;
  Rrun := abs(X2-X1);
  Rise := abs(Y2-Y1);
  IF Rrun > Rise THEN // travel along X, adjust Y per decision
    begin
      LineLen  := Rrun + 1;
      decision := (2 * Rise) - Rrun;
      dIncNeg  := 2 * Rise;
      dIncPos  := 2 * (Rise - Rrun);
      xIncNeg  := 1;
      xIncPos  := 1;
      yIncNeg  := 0;
      yIncPos  := 1;
    end
  ELSE            // travel along Y, adjust X per decision
    begin
      LineLen  := Rise + 1;
      decision := (2 * Rrun) - Rise;
      dIncNeg  := 2 * Rrun;
      dIncPos  := 2 * (Rrun - Rise);
      xIncNeg  := 0;
      xIncPos  := 1;
      yIncNeg  := 1;
      yIncPos  := 1;
    end;
  IF X1 > X2 THEN // going left to RBtn
    begin
      xIncNeg  := -xIncNeg;
      xIncPos  := -xIncPos;
    end;
  IF Y1 > Y2 THEN // going bottom to top
    begin
      yIncNeg  := -yIncNeg;
      yIncPos  := -yIncPos;
    end;
  X := 0; Y := 0;
  FOR N := 0 TO (LineLen DIV 2)-1 DO
    begin
      fFrame.PaintOne(X1+X, Y1+Y, ColorIdx, True);
      fFrame.PaintOne(X2-X, Y2-Y, ColorIdx, True);
      IF decision < 0 THEN
        begin
          decision := decision + dIncNeg;
          Inc(X, xIncNeg);
          Inc(Y, yIncNeg);
        end
      ELSE
        begin
          decision := decision + dIncPos;
          Inc(X, xIncPos);
          Inc(Y, yIncPos);
        end;
    end;
  IF Odd(LineLen) THEN // the final middle pixel
    fFrame.PaintOne(X1+X, Y1+Y, ColorIdx, True);
end;

procedure TDrawingTool.FrameOK;
//A precaution; call at start of each method that references fFrame
begin
  IF NOT Assigned(fFrame) THEN
    Raise(Exception.Create('TDrawingTool error; cannot draw '+
      'when frame is nil.'));
end;

procedure TDrawingTool.XorLines(OnOff: Boolean);
//Turn on/off the canvas settings needed for drawing XOR lines
begin
  FrameOK; //Verify that fFrame object is valid
  WITH fFrame.pbDraw.Canvas DO
    IF OnOff THEN
      begin
        Pen.Color   := clWhite;
        Pen.Style   := psDot;
        Pen.Mode    := pmXor;
        Brush.Style := bsClear;
      end
    ELSE
      begin
        Pen.Color := clBlack;
        Pen.Style := psSolid;
        Pen.Mode  := pmCopy;
      end;
end;

{ TDPencil }

procedure TDPencil.DrawStart(X, Y, fore, back: Integer;
  RBtn : Boolean; vFrame : TImageFrame);
//Initialize and draw a dot at starting position
begin
  Inherited;
  fFrame.UndoTool('Pencil');
  fFrame.PaintOne(X, Y, fFore, True);
  fDirty(Self);
end;

procedure TDPencil.DrawMove(X, Y: Integer; constrain: Boolean);
//Draw a line between current position and previous. This makes
//  drawing continuous even if the mouse is moving fast
begin
  FrameOK;
  DrawLine(fStartX, fStartY, X, Y, fFore);
  fStartX := X;
  fStartY := Y;
end;

procedure TDPencil.DrawEnd(X, Y: Integer);
//Draw a dot at ending position
begin
  FrameOK;
  fFrame.PaintOne(X, Y, fFore, True);
end;

{ TDDropper }

procedure TDDropper.DrawEnd(X, Y: Integer);
begin
  FrameOK;
  fFrame.ColorToPal(X, Y, NOT fRBtn);
end;

{ TDPaintCan }

procedure TDPaintCan.DrawEnd(X, Y: Integer);
//Recursive routine to fill a continuous region of the color at
//  (X,Y) with the Fore color
VAR ToFill : Integer;

  procedure FillPixel(vX, vY : Integer);
  begin
    IF vX < 0 THEN Exit; IF vX >= fFrame.Dimension THEN Exit;
    IF vY < 0 THEN Exit; IF vY >= fFrame.Dimension THEN Exit;
    IF fFrame.IcoCells[vX,vY] <> toFill THEN Exit;
    fFrame.PaintOne(vX,vY,fFore, True);
    FillPixel(vX-1, vY);
    FillPixel(vX+1, vY);
    FillPixel(vX, vY-1);
    FillPixel(vX, vY+1);
  end;

begin
  FrameOK;
  ToFill := fFrame.IcoCells[X,Y];
  IF ToFill = fFore THEN Exit;
  fFrame.UndoTool('Paint can');
  FillPixel(X,Y);
  fDirty(Self);
end;

{ TDrawingToolC }
procedure TDrawingToolC.SnapRect(X1, Y1: Integer; var X2, Y2: Integer);
//Given rectangle coordinates (X1,Y1) and (X2,Y2), adjust X2 and Y2
//  so the rectangle becomes a square
begin
  IF Abs(X2-X1) > Abs(Y2-Y1) THEN
    IF (X2>X1) XOR (Y2>Y1) THEN Y2 := Y1 + X1 - X2
    ELSE                        Y2 := Y1 + X2 - X1
  ELSE
    IF (X2>X1) XOR (Y2>Y1) THEN X2 := X1 + Y1 - Y2
    ELSE                        X2 := X1 + Y2 - Y1;
end;

function TDrawingToolC.CoordString(X, Y: Integer): String;
begin
  IF (fPrevX = MaxInt) OR (fPrevY = MaxInt) THEN
    Result := Inherited CoordString(X,Y)
  ELSE Result := Format('(%d,%d) -> (%d,%d)',
    [fStartX, fStartY, fPrevX, fPrevY]);
end;

{ TDLine }
procedure TDLine.SnapLine(X1, Y1: Integer; var X2, Y2: Integer);
//Given line coordinates (X1,Y1) and (X2,Y2), adjust X2 and Y2 so
// the line is horizontal, vertical, or 45 degrees
VAR factor : Double;
begin
  IF (X2=X1) OR (Y2=Y1) THEN Exit;
  factor := (X2-X1) / (Y2-Y1);
  IF Factor >= 2         THEN Y2 := Y1
  ELSE IF Factor <= -2   THEN Y2 := Y1
  ELSE IF Factor >= 1    THEN X2 := X1 + Y2 - Y1
  ELSE IF Factor <= -1   THEN X2 := X1 + Y1 - Y2
  ELSE IF Factor >= 0.5  THEN Y2 := Y1 + X2 - X1
  ELSE IF Factor <= -0.5 THEN Y2 := Y1 + X1 - X2
  ELSE                        X2 := X1;
end;

procedure TDLine.DrawMove(X, Y: Integer; constrain: Boolean);
begin
  FrameOK;
  XorLines(True);
  IF constrain THEN SnapLine(fStartX, fStartY, X, Y);
  WITH fFrame, pbDraw.Canvas DO
    begin
      // Clear previous XOR-line, if present
      IF (fPrevX < MaxInt) AND (fPrevY < MaxInt) THEN
        WITH Rect(CenterOf(fPrevX), CenterOf(fPrevY),
          CenterOf(fStartX), CenterOf(fStartY)) DO
          PolyLine([Point(Left, Top), Point(Right, Bottom)]);
      // Draw current XOR-line
      WITH Rect(CenterOf(X), CenterOf(Y),
        CenterOf(fStartX), CenterOf(fStartY)) DO
        PolyLine([Point(Left, Top), Point(Right, Bottom)]);
    end;
  XorLines(False);
  fPrevX := X;
  fPrevY := Y;
end;

procedure TDLine.DrawEnd(X, Y: Integer);
begin
  FrameOK;
  IF (fPrevX = MaxInt) OR (fPrevY = MaxInt) THEN Exit;
  DrawCancel;
  fFrame.UndoTool('Line');
  DrawLine(fStartX, fStartY, fPrevX, fPrevY, fFore);
  fDirty(Self);
end;

procedure TDLine.DrawCancel;
begin
  FrameOK;
  XorLines(True);
  // Clear previous XOR-line
  WITH fFrame, pbDraw.Canvas DO
    WITH Rect(CenterOf(fPrevX), CenterOf(fPrevY),
      CenterOf(fStartX), CenterOf(fStartY)) DO
        PolyLine([Point(Left, Top),
          Point(Right, Bottom)]);
  XorLines(False);
end;

{ TDRect }
function TDRect.RectAlign(fL, fT, fR, fB: Integer): TRect;
// Given two corners, return a rectangle in which left is
// less than right and top is less than bottom
begin
  IF fL < fR THEN begin Result.Left := fL; Result.Right  := fR; end
  ELSE            begin Result.Left := fR; Result.Right  := fL; end;
  IF fT < fB THEN begin Result.Top  := fT; Result.Bottom := fB; end
  ELSE            begin Result.Top  := fB; Result.Bottom := fT; end;
end;

constructor TDRect.Create(vDirty : TNotifyEvent);
begin
  Inherited Create(vDirty);
  fFill := False;
end;

procedure TDRect.DrawMove(X, Y: Integer; constrain: Boolean);
begin
  FrameOK;
  XorLines(True);
  IF constrain THEN SnapRect(fStartX, fStartY, X, Y);
  WITH fFrame, pbDraw.Canvas DO
    begin
      // Clear previous XOR-rect, if present
      IF (fPrevX < MaxInt) AND (fPrevY < MaxInt) THEN
        WITH RectAlign(CenterOf(fPrevX), CenterOf(fPrevY),
          CenterOf(fStartX), CenterOf(fStartY)) DO
            Rectangle(Left, Top, Right+Adjust, Bottom+Adjust);
      // Draw current XOR-rect
      WITH RectAlign(CenterOf(X), CenterOf(Y),
        CenterOf(fStartX), CenterOf(fStartY)) DO
          Rectangle(Left, Top, Right+Adjust, Bottom+Adjust);
    end;
  XorLines(False);
  fPrevX := X;
  fPrevY := Y;
end;

procedure TDRect.DrawEnd(X, Y: Integer);
VAR X1, X2, Y1, Y2, co, ro : Integer;
begin
  FrameOK;
  IF (fPrevX = MaxInt) OR (fPrevY = MaxInt) THEN Exit;
  DrawCancel;
  IF fFill THEN
    fFrame.UndoTool('Filled rectangle')
  ELSE fFrame.UndoTool('Rectangle');
  WITH RectAlign(fStartX, fStartY, fPrevX, fPrevY) DO
    begin
      X1 := Left; X2 := Right;
      Y1 := Top;  Y2 := Bottom;
    end;
  // Draw top and bottom lines
  FOR co := X1 TO X2 DO
    begin
      fFrame.PaintOne(co, Y1, fFore, True);
      fFrame.PaintOne(co, Y2, fFore, True);
    end;
  // Draw sides and, if needed, filled interior
  FOR ro := (Y1+1) TO (Y2-1) DO
    begin
      fFrame.PaintOne(X1, ro, fFore, True);
      fFrame.PaintOne(X2, ro, fFore, True);
      IF NOT fFill THEN Continue;
      FOR co := X1+1 TO X2-1 DO
        fFrame.PaintOne(co, ro, fBack, True);
    end;
  fDirty(Self);
end;

procedure TDRect.DrawCancel;
begin
  FrameOK;
  XorLines(True);
  // Clear previous XOR-rect
  WITH fFrame, pbDraw.Canvas DO
    WITH RectAlign(CenterOf(fPrevX), CenterOf(fPrevY),
      CenterOf(fStartX), CenterOf(fStartY)) DO
        Rectangle(Left, Top, Right+Adjust, Bottom+Adjust);
  XorLines(False);
end;

{ TDRectF }

constructor TDRectF.Create(vDirty : TNotifyEvent);
begin
  Inherited Create(vDirty);
  fFill := True;
end;

{ TDEllipse }

constructor TDEllipse.Create(vDirty : TNotifyEvent);
begin
  Inherited Create(vDirty);
  fFill := False;
end;

procedure TDEllipse.DrawMove(X, Y: Integer; constrain: Boolean);
begin
  FrameOK;
  XorLines(True);
  IF constrain THEN SnapRect(fStartX, fStartY, X, Y);
  WITH fFrame, pbDraw.Canvas DO
    begin
      // Clear previous XOR-ellipse, if present
      IF (fPrevX < MaxInt) AND (fPrevY < MaxInt) THEN
        Ellipse(CenterOf(fStartX-abs(fStartX-fPrevX)),
          CenterOf(fStartY-abs(fStartY-fPrevY)),
          CenterOf(fStartX+abs(fStartX-fPrevX))+Adjust,
          CenterOf(fStartY+abs(fStartY-fPrevY))+Adjust);
      // Draw current XOR-ellipse
      Ellipse(CenterOf(fStartX-abs(fStartX-X)),
        CenterOf(fStartY-abs(fStartY-Y)),
        CenterOf(fStartX+abs(fStartX-X))+Adjust,
        CenterOf(fStartY+abs(fStartY-Y))+Adjust);
    end;
  XorLines(False);
  fPrevX := X;
  fPrevY := Y;
end;

procedure TDEllipse.DrawEnd(X, Y: Integer);
// Draw an ellipse, calculating which points to draw by
// using a variation on Bresenham's algorithm
VAR
  xAxis, yAxis : Integer;
  xAxSq, yAxSq : Integer;
  dec1, dec2   : Integer; // decision variables

  procedure Pixel4(vX, vY : Integer);
  begin
    fFrame.PaintOne(fStartX - vX, fStartY + vY, fFore, True);
    fFrame.PaintOne(fStartX + vX, fStartY + vY, fFore, True);
    fFrame.PaintOne(fStartX - vX, fStartY - vY, fFore, True);
    fFrame.PaintOne(fStartX + vX, fStartY - vY, fFore, True);
  end;

  procedure DoFill(vX, vY : Integer);
  VAR xL, xR, co : Integer;
  begin
    IF NOT fFill THEN
      begin
        fFill := False;
        Exit;
      end;
    IF vY = yAxis THEN Exit;
    xL := fStartX - vX + 1;
    IF xL < 0 THEN xL := 0;
    xR := fStartX + vX - 1;
    IF xR >= fFrame.Dimension THEN xR := fFrame.Dimension-1;
    FOR co := xL TO xR DO
      begin
        fFrame.PaintOne(co, fStartY+vY, fBack, True);
        fFrame.PaintOne(co, fStartY-vY, fBack, True);
      end;
  end;

begin
  FrameOK;
  IF (fPrevX = MaxInt) OR (fPrevY = MaxInt) THEN Exit;
  DrawCancel;
  IF fFill THEN
    fFrame.UndoTool('Filled ellipse')
  ELSE fFrame.UndoTool('Ellipse');
  xAxis := Abs(fStartX - fPrevX);
  yAxis := Abs(fStartY - fPrevY);
  xAxSq   := xAxis*xAxis; // precalculate x - axis squared
  yAxSq   := yAxis*yAxis; // precalculate y - axis squared
  X       := 0;
  y       := yAxis;
  dec1 := xAxSq * (1 - 2 * yAxis) + 2 * yAxSq;
  dec2 := yAxSq - 2 * xAxSq * (2 * yAxis - 1);
  Pixel4(x,y);
  WHILE y > 0 DO
    begin
      IF dec1 < 0 THEN
        begin
          Inc(dec1, 2 * yAxSq * (2*x + 3));
          Inc(dec2, 4 * yAxSq * (x + 1));
          Inc(X);
        end
      ELSE IF dec2 < 0 THEN
        begin
          Inc(dec1, 2 * yAxSq * (2*x + 3) - 4 * xAxSq * (y - 1));
          Inc(dec2, 4 * yAxSq * (x + 1) - 2 * xAxSq * (2*y - 3));
          Inc(X);
          Dec(Y);
          DoFill(X,Y);
        end
      ELSE
        begin
          Dec(dec1, 4 * xAxSq * (y - 1));
          Dec(dec2, 2 * xAxSq * (2*y - 3));
          Dec(Y);
          DoFill(X,Y);
        end;
      Pixel4(x,y);
    end;
  fDirty(Self);
end;

procedure TDEllipse.DrawCancel;
begin
  FrameOK;
  XorLines(True);
  // Clear previous XOR-ellipse
  WITH fFrame, pbDraw.Canvas DO
    Ellipse(CenterOf(fStartX-abs(fStartX-fPrevX)),
      CenterOf(fStartY-abs(fStartY-fPrevY)),
      CenterOf(fStartX+abs(fStartX-fPrevX))+Adjust,
      CenterOf(fStartY+abs(fStartY-fPrevY))+Adjust);
  XorLines(False);
end;

{ TDEllipseF }

constructor TDEllipseF.Create(vDirty : TNotifyEvent);
begin
  Inherited Create(vDirty);
  fFill := True;
end;

{ TDChooser }

procedure TDChooser.DrawStart(X, Y, fore, back: Integer; RBtn: Boolean;
  vFrame: TImageFrame);
begin
  inherited;
  fOrgRect := fFrame.FocusRect;
  fFrame.FocusRect := NoRect;
end;

procedure TDChooser.DrawEnd(X, Y: Integer);
begin
  FrameOK;
  IF (fPrevX = MaxInt) OR (fPrevY = MaxInt) THEN Exit;
  DrawCancel;
  WITH RectAlign(fStartX, fStartY, fPrevX, fPrevY) DO
    fFrame.FocusRect := Rect(Left, Top, Right, Bottom);
  fFrame.UndoWRect(fOrgRect, tfSelect);
  fFrame.FocusCellsInit;
end;

{ TDText }

procedure TDText.DrawEnd(X, Y: Integer);
VAR
  TB               : TBitmap;
  co, ro, co1, ro1 : Integer;
  S                : String;

begin
  FrameOK;
  IF TextForm.ShowModal <> idOK THEN Exit;
  S  := TextForm.ebSample.Text;
  IF S = '' THEN Exit;
  TB := TBitmap.Create;
  try
    TB.Width              := 48;
    TB.Height             := 48;
    TB.Canvas.Font        := TextForm.fFD.Font;
    TB.Canvas.Brush.Color := clWhite;
    TB.Canvas.Font.Color  := clBlack;
    TB.Canvas.TextOut(0, 0, S);
    fFrame.UndoTool('Text');
    //co1 is the first column that contains pixels
    co1 := -1;
    FOR co := 0 TO 48 DO
      begin
        FOR ro := 0 TO 48 DO
          IF TB.Canvas.Pixels[co,ro] = clBlack THEN
            begin
              co1 := co;
              Break;
            end;
          IF co1 >= 0 THEN break;
      end;
    IF co1 < 0 THEN co1 := 0;
    //ro1 is the first row that contains pixels
    ro1 := -1;
    FOR ro := 0 TO 48 DO
      begin
        FOR co := 0 TO 48 DO
          IF TB.Canvas.Pixels[co,ro] = clBlack THEN
            begin
              ro1 := ro;
              Break;
            end;
          IF ro1 >= 0 THEN Break;
      end;
    IF ro1 < 0 THEN ro1 := 0;
    for co := co1 TO 48 DO
      for ro := ro1 TO 48 DO
        IF TB.Canvas.Pixels[co,ro] = clBlack THEN
          fFrame.PaintOne(X+co-co1, Y+ro-ro1, fFore, True);
  finally
    TB.Free;
  end;
end;

{ TDMover }

procedure TDMover.DrawStart(X, Y, fore, back: Integer;
  RBtn: Boolean; vFrame: TImageFrame);
begin
  inherited;
  IF fFrame.PtInCoRect(fFrame.FocusRect, X, Y) THEN
    WITH fFrame.FocusRect DO
      begin
        fValid := True;
        fOrgRect := fFrame.FocusRect;
        WITH FFrame.FocusRect DO
          begin
            fPrevX := Left;
            fPrevY := Top;
          end;
      end
    ELSE fValid := False;
end;

procedure TDMover.DrawMove(X, Y: Integer; constrain: Boolean);
begin
  FrameOK;
  IF fValid THEN
    begin
      fFrame.FocusCellsMove(X-fStartX+fPrevX, Y-fStartY+fPrevY);
      fDirty(Self);
    end;
end;

procedure TDMover.DrawEnd(X, Y : Integer);
begin
  FrameOK;
  //Only generate an Undo event if the rect moved
  WITH fFrame.FocusRect DO
    IF (fOrgRect.Left = Left) AND (fOrgRect.Top = Top) THEN
      Exit;
  fFrame.UndoWRect(fOrgRect, tfMove);
end;

procedure TDMover.DoneUsing;
begin
  IF fValid AND (fFrame <> nil) THEN
    fFrame.FocusCellsDrop;
end;

end.



