unit icoObj;

// Copyright  2000 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

interface
USES Windows, SysUtils, Classes, palconsu;

type
  TIconDirEntry = packed record
  {.ICO icon directory contains one or more of these}
    bWidth,
    bHeight,
    bColorCount,
    bReserved     : Byte;
    wPlanes,
    wBitCount     : Word;
    dwBytesInRes,
    dwImageOffset : LongInt;
  End;

  TIconHeader = packed record
  {structure of .ICO header. actual number of idEntries
   is determined by idCount}
    idReserved : Word;
    idType     : Word;
    idCount    : Word;
  END;

IDEObject = class(TObject)
private
  fIDE     : TIconDirEntry;
  fPixels  : PByteArray;
  fPixSize : Integer;
  function GetColorCount     : Integer;
  function GetHeight         : Integer;
  function GetIcoCells       : TCellsType;
  function GetImageCode      : Integer;
  function GetImageString    : String;
  function GetPaletteCount   : Integer;
  function GetPaletteColors  : TPalArray;
  function GetWidth          : Integer;
  function NthPixel(X : PByteArray; N, bits : Integer) : Integer;
  procedure SetNthPixel(X : PByteArray; N, bits, value : Integer);
public
  constructor Create; overload;
  constructor Create(Dimension, ColorCode : Integer;
    vCells : TCellsType; vPal : TPalArray); overload;
  destructor Destroy; override;
  procedure LoadFromStream(Stream: TStream); virtual;
  procedure SaveIDEToStream(Stream: TStream); virtual;
  procedure SavePixToStream(Stream: TStream); virtual;

  property ColorCount    : Integer    read GetColorCount;
  property Height        : Integer    read GetHeight;
  property IcoCells      : TCellsType read GetIcoCells;
  property ImageCode     : Integer    read GetImageCode;
  property ImageString   : String     read GetImageString;
  property PaletteColors : TPalArray  read GetPaletteColors;
  property PaletteCount  : Integer    read GetPaletteCount;
  property Width         : Integer    read GetWidth;
end;

IcoFileObj = class(TObject)
private
  fHeader : TIconHeader;
  fIDEs   : TList;
  function GetIDEs(Index: Integer): IDEObject;
  procedure ClearIDEs;
public
  constructor Create;
  destructor Destroy; override;
  procedure LoadFromStream(Stream: TStream); virtual;
  procedure SaveToStream(Stream: TStream); virtual;
  procedure LoadFromFile(const FileName: string); virtual;
  procedure SaveToFile(const FileName: string); virtual;
  function IDEMatching(D, Colors : Integer) : IDEObject;
  function ValidImages : Boolean;
  procedure AddImage(Dimension, ColorCode : Integer;
    vCells : TCellsType; vPal : TPalArray);
  property IDEs[Index : Integer] : IDEObject read GetIDEs;
end;


implementation

type
  icoBITMAPINFOHEADER = packed record
    // This is just a BITMAPINFOHEADER. However, when it
    // appears within an icon, the fields that have "0"
    // appended to their names must be set to 0.
    biSize           : DWORD;
    biWidth          : Longint;
    biHeight         : Longint;
    biPlanes         : Word;
    biBitCount       : Word;
    biCompression0   : DWORD;   // must be 0
    biSizeImage      : DWORD;
    biXPelsPerMeter0 : Longint; // must be 0
    biYPelsPerMeter0 : Longint; // must be 0
    biClrUsed0       : DWORD;   // must be 0
    biClrImportant0  : DWORD;   // must be 0
  end;

{ IDEObject }

constructor IDEObject.Create;
begin
  Inherited Create;
  FillChar(fIDE, SizeOf(fIDE), 0);
  fPixels  := nil;
  fPixSize := 0;
end;

constructor IDEObject.Create(Dimension, ColorCode: Integer;
  vCells: TCellsType; vPal : TPalArray);
VAR
  icHeader         : icoBITMAPINFOHEADER;
  XORBits, ANDBits : PByteArray;
  PalSize          : Integer;
  XORsize, ANDsize : Integer;
  XRow, ARow       : Integer;
  theOfs           : Integer;
  co, ro, N        : Integer;
  RQs              : ARRAY[0..255] OF TRGBQuad;
begin
  Inherited Create;
  //Initialize icHeader. All fields 0 except those specified
  FillChar(icHeader, SizeOf(icHeader), 0);
  icHeader.biSize       := SizeOf(icHeader);
  icHeader.biWidth      := Dimension;
  icHeader.biHeight     := 2*Dimension;
  icHeader.biPlanes     := 1;
  CASE colorCode OF
    1 :  icHeader.biBitCount := 4;
    2 :  icHeader.biBitCount := 1;
    ELSE icHeader.biBitCount := 8;
  END;
  //icHeader.biSizeImage is initialized after other calculations
  //Initialize fIDE. Its dwBytesInRes field is initialized
  //  after the other calculations
  FillChar(fIDE, SizeOf(fIDE), 0);
  fIDE.bWidth             := Dimension;
  fIDE.bHeight            := Dimension;
  fIDE.wBitCount          := icHeader.biBitCount;
  CASE colorCode OF
    1 :  fIDE.bColorCount := 16;
    2 :  fIDE.bColorCount := 2;
    ELSE fIDE.bColorCount := 0;
  END;
  fIDE.bReserved          := 0;
  fIDE.wPlanes            := 1;
  fIDE.dwImageOffset      := 0;

  // Set palette size
  CASE colorCode OF
    1 :  PalSize             :=  16*SizeOf(TRGBQuad);
    2 :  PalSize             :=   2*SizeOf(TRGBQuad);
    ELSE PalSize             := 256*SizeOf(TRGBQuad);
  END;
  //XRow is the actual number of bytes per scanline in the XOR bitmap
  XRow := icHeader.biWidth * icHeader.biBitCount DIV 8;
  IF XRow MOD 4 <> 0 THEN XRow := XRow + 4 - (XRow MOD 4);
  XRow := XRow * 8 DIV icHeader.biBitCount;
  //ARow is the actual number of bytes per scanline in the AND bitmap
  ARow := icHeader.biWidth DIV 8;
  IF ARow MOD 4 <> 0 THEN ARow := ARow + 4 - (ARow MOD 4);
  ARow := ARow * 8;
  // Calculate size of XOR and AND bitmaps
  XORsize := ((icHeader.biHeight DIV 2) * XRow * icHeader.biBitCount)
    DIV 8;
  ANDsize := ((icHeader.biHeight DIV 2) * ARow) DIV 8;
  icHeader.biSizeImage  := XORsize + ANDsize;
  fPixSize              := SizeOf(icHeader) + PalSize +
    XORsize + ANDsize;
  fIDE.dwBytesInRes     := fPixSize;
  GetMem(fPixels, fPixSize);
  FillChar(fPixels^, fPixSize, 0);
  TheOfs := 0; //Initialize offset for filling fPixels
  Move(icHeader, fPixels^[theOfs], SizeOf(icHeader));
  Inc(theOfs, SizeOf(icHeader));
  FillChar(RQs, SizeOf(RQs), 0);
  FOR N := 0 TO PalSize DIV SizeOf(TRGBQuad) DO
    begin
      RQs[N].rgbBlue  := GetBValue(vPal[N]);
      RQs[N].rgbGreen := GetGValue(vPal[N]);
      RQs[N].rgbRed   := GetRValue(vPal[N]);
    end;
  Move(RQs, fPixels^[theOfs], PalSize);
  Inc(theOfs, PalSize);
  XORBits := @fPixels^[theOfs];
  Inc(theOfs, XORSize);
  ANDbits := @fPixels^[theOfs];
  FOR ro := 0 TO icHeader.biWidth-1 DO
    FOR co := 0 TO icHeader.biWidth-1 DO
      begin
        CASE vCells[co,ro] OF
          palTrans : begin //AND=white, XOR=black
            SetNthPixel(ANDBits,
              (icHeader.biWidth-1-ro)*ARow+co,
              1, 1);
            SetNthPixel(XORBits,
              (icHeader.biWidth-1-ro)*XRow+co,
              icHeader.biBitCount, 0);
          end;
          palInvrt : begin //AND=white, XOR=white
            SetNthPixel(ANDBits,
              (icHeader.biWidth-1-ro)*ARow+co,
              1, 1);
            SetNthPixel(XORBits,
              (icHeader.biWidth-1-ro)*XRow+co,
              icHeader.biBitCount, $FF);
          end;
          ELSE //AND=black, XOR=image color
            SetNthPixel(XORBits,
              (icHeader.biWidth-1-ro)*XRow+co,
              icHeader.biBitCount, vCells[co,ro]);
        end;
      end;
end;

destructor IDEObject.Destroy;
begin
  IF Assigned(fPixels) THEN FreeMem(fPixels);
  inherited;
end;

procedure IDEObject.LoadFromStream(Stream: TStream);
VAR
  strmpos  : Integer;
  icHeader : icoBITMAPINFOHEADER;
begin
  Stream.Read(fIDE, SizeOf(fIDE));
  strmpos := Stream.Position;
  Stream.Seek(fIDE.dwImageOffset, soFromBeginning);
  IF Assigned(fPixels) THEN FreeMem(fPixels);
  fPixSize := fIDE.dwBytesInRes;
  GetMem(fPixels, fPixSize);
  Stream.Read(fPixels^, fPixSize);
  Stream.Seek(strmpos, sofromBeginning);
  Move(fPixels^, icHeader, SizeOf(icHeader));
  CASE icHeader.biBitCount OF
    1 :  fIDE.bColorCount := 2;
    4 :  fIDE.bColorCount := 16;
    ELSE fIDE.bColorCount := 0;
  END;
end;

procedure IDEObject.SaveIDEToStream(Stream: TStream);
begin
  Stream.Write(fIDE, SizeOf(fIDE));
end;

procedure IDEObject.SavePixToStream(Stream: TStream);
begin
  fIDE.dwImageOffset := Stream.Position;
  Stream.Write(fPixels^, fPixSize);
end;

function IDEObject.GetColorCount: Integer;
begin
  Result := fIDE.bColorCount;
  IF Result = 0 THEN Result := 256;
end;

function IDEObject.GetHeight: Integer;
begin
  Result := fIDE.bHeight;
end;

function IDEObject.GetIcoCells: TCellsType;
VAR
  icHeader         : icoBITMAPINFOHEADER;
  XORBits, ANDBits : PByteArray;
  XORsize, ANDsize : Integer;
  XRow, ARow       : Integer;
  theOfs           : Integer;
  co, ro           : Integer;
begin
  IF ImageCode < 0 THEN
    Raise(Exception.Create('Attempt to read pixels of an '+
      'invalid image type'));
  FillChar(Result, SizeOf(Result), 0);
  Move(fPixels^, icHeader, SizeOf(icHeader));
  theOfs := SizeOf(icHeader);
  CASE icHeader.biBitCount OF
    1 :  Inc(theOfs,   2*SizeOf(TRGBQuad));
    4 :  Inc(theOfs,  16*SizeOf(TRGBQuad));
    ELSE Inc(theOfs, 256*SizeOf(TRGBQuad));
  END;
  //XRow is the actual number of bytes per scanline in the XOR bitmap
  XRow := icHeader.biWidth * icHeader.biBitCount DIV 8;
  IF XRow MOD 4 <> 0 THEN
    XRow := XRow + 4 - (XRow MOD 4);
  XRow := XRow * 8 DIV icHeader.biBitCount;
  //ARow is the actual number of bytes per scanline in the AND bitmap
  ARow := icHeader.biWidth DIV 8;
  IF ARow MOD 4 <> 0 THEN
    ARow := ARow + 4 - (ARow MOD 4);
  ARow := ARow * 8;
  XORbits := @fPixels^[theOfs];
  XORsize := ((icHeader.biHeight DIV 2) * XRow * icHeader.biBitCount)
    DIV 8;
  Inc(theOfs, XORsize);
  ANDbits := @fPixels^[theOfs];
  ANDsize := ((icHeader.biHeight DIV 2) * ARow) DIV 8;
  Inc(theOfs, ANDsize);
  IF theOfs <> fPixSize THEN
    Raise(Exception.Create('GetIcoCells - error in data size'));
  FOR ro := 0 TO icHeader.biWidth-1 DO
    FOR co := 0 TO icHeader.biWidth-1 DO
      begin
        Result[co,ro] := NthPixel(XORBits,
          (icHeader.biWidth-1-ro)*XRow+co,
          icHeader.biBitCount);
        // IF the AND mask has a white pixel...
        IF NthPixel(ANDBits, (icHeader.biWidth-1-ro)*ARow+co,
          1) = 1 THEN
          begin
            IF Result[co,ro] = (1 SHL icHeader.biBitCount) - 1 THEN
              Result[co,ro] := palInvrt
            ELSE IF Result[co,ro] = 0 THEN
              Result[co,ro] := palTrans;
          end;
      end;
end;

function IDEObject.GetImageCode: Integer;
begin
  Result := -1;
  IF fIDE.bWidth <> fIDE.bHeight THEN Exit;
  CASE ColorCount OF
    256 : CASE fIDE.bWidth OF
      48 : Result := 0;
      32 : Result := 1;
      16 : Result := 2;
    END;
    16  : CASE fIDE.bWidth OF
      48 : Result := 3;
      32 : Result := 4;
      16 : Result := 5;
    END;
    2   : CASE fIDE.bWidth OF
      48 : Result := 6;
      32 : Result := 7;
      16 : Result := 8;
    END;
  END;
end;

function IDEObject.GetImageString: String;
begin
  Result := Format('%dx%d, %d colors', [fIDE.bWidth,
    fIDE.bHeight, ColorCount]);
end;

function IDEObject.GetPaletteColors: TPalArray;
VAR
  icHeader : icoBITMAPINFOHEADER;
  RGBQuads : ARRAY[0..255] OF TRGBQuad;
  Count, N : Integer;
begin
  Move(fPixels^, icHeader, SizeOf(icHeader));
  CASE icHeader.biBitCount OF
    1 :  Count := 2;
    4 :  Count := 16;
    ELSE Count := 256;
  END;
  FillChar(Result, SizeOf(Result), 0);
  FillChar(RGBQuads, SizeOf(RGBQuads), 0);
  Move(fPixels^[SizeOf(icHeader)], RGBQuads, SizeOf(TRGBQuad)*Count);
  FOR N := 0 TO Count-1 DO
    WITH RGBQuads[N] DO
      Result[N] := RGB(rgbRed, rgbGreen, rgbBlue);
end;

function IDEObject.GetPaletteCount: Integer;
VAR icHeader : icoBITMAPINFOHEADER;
begin
  Move(fPixels^, icHeader, SizeOf(icHeader));
  CASE icHeader.biBitCount OF
    1 :  Result :=   2;
    4 :  Result :=  16;
    ELSE Result := 256;
  END;
end;

function IDEObject.GetWidth: Integer;
begin
  Result := fIDE.bWidth;
end;

function IDEObject.NthPixel(X : PByteArray; N, bits : Integer) :
  Integer;
// Returns the Nth pixel in an array of 1, 4, or 8 bit pixel
// values. N is linear, calculated from the row and column
// positions by the calling routine.
begin
  CASE bits OF
    1 : Result := (X^[N DIV 8] SHR ((7-(N MOD 8))*1)) AND 1;
    4 : Result := (X^[N DIV 2] SHR ((1-(N MOD 2))*4)) AND $F;
    ELSE Result := X^[N];
  end;
end;

procedure IDEObject.SetNthPixel(X : PByteArray; N, bits,
  value : Integer);
VAR B, Mask : Byte;
// Sets the Nth pixel in an array of 1-, 4-, or 8-bit pixel
// values. N is linear, calculated from the row and column
// positions by the calling routine.
begin
  CASE bits OF
    1 : begin
      B    := X^[N DIV 8];
      Mask := (NOT (1 SHL ((7-(N MOD 8))*1)));
      B    := (B AND Mask) OR ((Value AND 1) SHL ((7-(N MOD 8))*1));
      X^[N DIV 8] := B;
    end;
    4 : begin
      B := X^[N DIV 2];
      IF Odd(N) THEN
        B := (B AND $F0) OR (Value AND $F)
      ELSE B := (B AND $F) OR ((Value  AND $F) SHL 4);
      X^[N DIV 2] := B;
    end;
    ELSE X^[N] := Value;
  end;
end;

{ IcoFileObj }

constructor IcoFileObj.Create;
begin
  Inherited Create;
  FillChar(fHeader, SizeOf(fHeader), 0);
  fIDEs := TList.Create;
end;

destructor IcoFileObj.Destroy;
begin
  ClearIDEs;
  fIDEs.Free;
  inherited;
end;

procedure IcoFileObj.LoadFromStream(Stream: TStream);
VAR
  N : Integer;
  I : IDEObject;
begin
  Stream.Read(fHeader, SizeOf(fHeader));
  IF fHeader.idReserved <> 0 THEN
    Raise(Exception.Create('IcoFileObj.LoadFromStream - not a valid '+
      '.ICO file.'));
  IF fHeader.idType     <> 1 THEN
    Raise(Exception.Create('IcoFileObj.LoadFromStream - not a valid '+
      '.ICO file.'));
  IF fHeader.idCount < 1 THEN
    Raise(Exception.Create('IcoFileObj.LoadFromStream - not a valid '+
      '.ICO file.'));
  ClearIDEs;
  FOR N := 0 TO fHeader.idCount-1 DO
    begin
      I := IDEObject.Create;
      I.LoadFromStream(Stream);
      fIDEs.Add(I);
    end;
end;

procedure IcoFileObj.SaveToStream(Stream: TStream);
VAR
  N, IDEsStart : Integer;
begin
  IF fIDEs.Count < 1 THEN
    Raise(Exception.Create('IcoFileObj not property initialized'));
  fHeader.idReserved := 0;
  fHeader.idType     := 1;
  fHeader.idCount    := fIDEs.Count;
  Stream.Write(fHeader, SizeOf(fHeader));
  IDEsStart := Stream.Position;
  FOR N := 0 TO fIDEs.Count-1 DO
    IDEObject(fIDEs[N]).SaveIDEToStream(Stream);
  FOR N := 0 TO fIDEs.Count-1 DO
    IDEObject(fIDEs[N]).SavePixToStream(Stream);
  Stream.Position := IDEsStart;
  FOR N := 0 TO fIDEs.Count-1 DO
    IDEObject(fIDEs[N]).SaveIDEToStream(Stream);
end;

procedure IcoFileObj.LoadFromFile(const FileName: string);
VAR Stream: TStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead OR
    fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure IcoFileObj.SaveToFile(const FileName: string);
VAR Stream: TStream;
begin
  FileClose(FileCreate(Filename));
  Stream := TFileStream.Create(FileName, fmOpenReadWrite);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

function IcoFileObj.IDEMatching(D, Colors: Integer): IDEObject;
VAR N : Integer;
begin
  Result := nil;
  FOR N := 0 TO fIDEs.Count-1 DO
    WITh IDEObject(fIDEs[N]) DO
      IF (ImageCode >= 0) AND (Width=D) AND (ColorCount=Colors) THEN
        begin
          Result := IDEObject(fIDEs[N]);
          Break;
        end;
end;

function IcoFileObj.ValidImages : Boolean;
const
  ImStrings : ARRAY[0..8] OF String =
    ('48x48, 256-color',
     '32x32, 256-color',
     '16x16, 256-color',
     '48x48, 16-color',
     '32x32, 16-color',
     '16x16, 16-color',
     '48x48, 2-color',
     '32x32, 2-color',
     '16x16, 2-color');

VAR
  N, M, code : Integer;
  TS         : TStringList;
  Counts     : ARRAY[0..8] OF Integer;
  err        : Boolean;
begin
  Err := False;
  FillChar(Counts, SizeOf(Counts), 0);
  TS := TStringList.Create;
  try
    FOR N := 0 TO fIDEs.Count-1 DO
      begin
        code := IDEObject(fIDEs[N]).ImageCode;
        IF Code IN [0..8] THEN Inc(Counts[code])
        ELSE
          begin
            TS.Add(IDEObject(fIDEs[N]).ImageString +
              #9'- NON-STANDARD');
            err := True;
          end;
      end;
    IF NOT Err THEN
      FOR N := 0 TO 8 DO
        IF Counts[N] > 1 THEN
          begin
            err := True;
            Break;
          end;
    IF err THEN
      begin
        Result := False;
        FOR N := 0 TO 8 DO
          IF Counts[N] > 0 THEN
            begin
              TS.Add(ImStrings[N] + #9'- ok');
              FOR M := 2 TO Counts[N] DO
                TS.Add(ImStrings[N] + #9'- DUPLICATE');
            end;
        MessageBox(0, PChar('This icon contains some images that ar'+
          'e not valid:'#13#10#13#10+TS.Text+#13#10#13#10+'You may '+
          'view the valid images in IconEdit32 without any problem.'+
          #13#10'However, if you save this icon, the invalid images'+
          ' will be discarded.'), 'IconEdit32 - WARNING',
          MB_OK OR MB_ICONWARNING);
      end
    ELSE Result := True;
  finally
    TS.Free;
  end;
end;

procedure IcoFileObj.AddImage(Dimension, ColorCode: Integer;
  vCells: TCellsType; vPal: TPalArray);
begin
  fIDEs.Add(IDEObject.Create(Dimension, ColorCode,
    vCells, vPal));
end;

function IcoFileObj.GetIDEs(Index: Integer): IDEObject;
begin
  IF (Index < 0) OR (Index >= fIDEs.Count) THEN
    Raise(Exception.Create('Invalid index for IDE array'));
  Result := fIDEs[Index];
end;

procedure IcoFileObj.ClearIDEs;
VAR N : Integer;
begin
  FOR N := fIDEs.Count-1 DOWNTO 0 DO
    begin
      IDEObject(fIDEs[N]).Free;
      fIDEs.Delete(N);
    end;
end;

end.
