unit imFrameU;

// Copyright  2000 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, palfrm2u, palconsu;

const
  NoRect : TRect = (Left:0; Top:0; Right:-1; Bottom:-1);
type
  TCellEvent = procedure(Sender : TObject; X, Y,
    Stat : Integer) of object;
  TTransform = (tfShiftLeft, tfShiftRight, tfShiftUp, tfShiftDown,
    tfRotateLeft, tfRotateRight, tfMirror, tfFlip, tfSelect, tfMove,
    tfDrop, tfClear, tfShift, tfPaste, tfColor, tfTool);

  TSavedCell = packed record
    fCo, fRo : Byte;
    fColor   : SmallInt;
  end;

  TUndoObj = class(TObject)
  // Stores Undo information after use of a tool or action
  private
    fList      : TList;      // list of cells and previous values
    fTransform : TTransform; // action-type transformation
    fTool      : String;     // name of tool used
    fRect      : TRect;      // focus rect
    function GetUndoName: String;
    class function UndoNameOf(fTransform : TTransform) : String;
  public
    constructor Create(const vTool : String); overload;
    constructor Create(vTransform : TTransform;
      const vTool : String); overload;
    destructor Destroy; override;
    procedure AddCell(co, ro, colr : Integer);
    procedure RemCell(VAR co, ro, colr : Integer);
    property UndoName : String read GetUndoName;
    property FocusRect : TRect read fRect write fRect;
  end;

  TImageFrame = class(TFrame)
    pbDraw: TPaintBox;
    procedure FrameResize(Sender: TObject);
    procedure FrameMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbDrawMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbDrawMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbDrawMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbDrawPaint(Sender: TObject);
  private
    { Private declarations }
    fAdj        : Integer;    //adjustment for rects and ellipses
    fCellAll    : Integer;    //size of all cells together
    fCellHalf   : Integer;    //size of half a cell
    fCellOne    : Integer;    //size of one cell
    fCells      : TCellsType; //array of color indices for cells
    fCellsM     : TCellsType; //array of color indices for selection
    fColorCount : Integer;    //2, 16, or 256
    fDimension  : Integer;    //16, 32, or 48
    fDMax       : Integer;    //one less than fDimension
    fECellDown  : TCellEvent; //sent on mouse-down
    fECellMove  : TCellEvent; //send on mouse-move w/ changed cell
    fECellUp    : TCellEvent; //sent on mouse-up
    fExists     : Boolean;    //true if image exists in current icon
    fFoRect     : TRect;      //focus rectangle
    fGotDown    : Integer;    //2=right btn, 1=left btn, 0=no btn
    fMouseX     : Integer;    //Previous cell position, X
    fMouseY     : Integer;    //Previous cell position, Y
    fMoving     : Boolean;    //True if moving selection
    fPal        : TPalFrame2; //color palette frame
    fPalColors  : TPalArray;  //array of palette colors
    fUndoList   : TList;      //list of undo objects for this image
    fUndoing    : Boolean;    //true during undo
    //property Get/Set methods
    procedure SetCellsM(const Value: TCellsType);
    function  GetColorCode: Integer;
    procedure SetColorCode(const Value: Integer);
    function  GetDimension: Integer;
    procedure SetDimension(const Value: Integer);
    procedure SetFocusRect(const Value: TRect);
    function  GetImageBlank: Boolean;
    function  GetPixelCol(X, Y: Integer): SmallInt;
    function  GetUndoName: String;
    //other private methods
    procedure CalcCells;
    function  CellFmPixel(X, Y: Integer; VAR cX, cY: Integer):
      Boolean;
    procedure ShowFocusRect(Show : Boolean);
    function  ValidCell(vCo, vRo: Integer): Boolean;
  public
    { Public declarations }
    procedure ActChangeCol;
    procedure ActShift(typ : Integer);
    procedure ActTransform(typ : Integer);
    function  CenterOf(I : Integer) : Integer;
    procedure ClearCells;
    procedure ColorToPal(co, ro : Integer; isFore : Boolean);
    function  ColString(cX, cY : Integer) : String;
    procedure FocusCellsPreClear;
    procedure FocusCellsClear;
    procedure FocusCellsDrop;
    procedure FocusCellsInit;
    procedure FocusCellsGet;
    procedure FocusCellsMove(dX, dY : Integer);
    procedure JustPaint(co, ro, colr :Integer);
    procedure PaintOne(co, ro, colr :Integer; un : Boolean);
    function  PtInCoRect(R : TRect; X, Y: Integer): Boolean;
    procedure Setup(D, M, U : TCellEvent; P : TPalFrame2);
    function  UndoAble : Boolean;
    procedure UndoDoIt;
    procedure UndoMaxReached;
    function UndoTool(const vTool : String) : TUndoObj;
    function UndoTrans(vTransform : TTransform) : TUndoObj;
    function UndoWRect(R : TRect; vTransform : TTransform) : TUndoObj;
    // properties
    property Adjust     : Integer    read  fAdj;
    property ColorCode  : Integer    read  GetColorCode
                                     write SetColorCode default 2;
    property ColorCount : Integer    read  fColorCount;
    property Dimension  : Integer    read  GetDimension
                                     write SetDimension default 32;
    property FocusRect  : TRect      read  fFoRect
                                     write SetFocusRect;
    property IcoCells   : TCellsType read  fCells
                                     write fCells;
    property ImageBlank : Boolean    read  GetImageBlank;
    property ImageExists : Boolean   read  fExists
                                     write fExists;
    property PalColors  : TPalArray  read  fPalColors
                                     write fPalColors;
    property SelCells   : TCellsType read  fCellsM
                                     write SetCellsM;
    property UndoName   : String     read  GetUndoName;
    property PixelCol [X : Integer; Y : Integer] : SmallInt
                                     read GetPixelCol;
  end;

implementation

const UNDOMAX = 64;
{$R *.DFM}

{ TUndoObj }

constructor TUndoObj.Create(const vTool : String);
begin
  Inherited Create;
  fList      := TList.Create;
  fTransform := tfTool;
  fTool      := vTool;
  fRect      := NoRect;
end;

constructor TUndoObj.Create(vTransform: TTransform;
  const vTool : String);
begin
  Create(vTool); // NOT inherited - call to other overloaded Create
  fTransform := vTransform;
end;

destructor TUndoObj.Destroy;
begin
  fList.Free;
  inherited;
end;

procedure TUndoObj.AddCell(co, ro, colr: Integer);
// A TSavedCell is exactly 4 bytes in size. We add a TSavedCell
// to the list by typecasting it as a pointer
VAR TS : TSavedCell;
begin
  TS.fCo    := Co;
  TS.fRo    := Ro;
  TS.fColor := Colr;
  fList.Add(Pointer(TS));
end;

procedure TUndoObj.RemCell(var co, ro, colr: Integer);
// Remove the last cell from the list and return its data in
// the VAR parameters. If the list is empty, return all -1
VAR TS : TSavedCell;
begin
  IF fList.Count = 0 THEN
    begin
      co   := -1;
      ro   := -1;
      colr := -1;
    end
  ELSE
    begin
      Pointer(TS) := fList[fList.Count-1];
      fList.Delete(fList.Count-1);
      co   := TS.fCo;
      ro   := TS.fRo;
      colr := TS.fColor;
    end;
end;

class function TUndoObj.UndoNameOf(fTransform : TTransform) : String;
begin
  CASE fTransform OF
    tfShiftLeft   : Result := 'Shift left';
    tfShiftRight  : Result := 'Shift right';
    tfShiftUp     : Result := 'Shift up';
    tfShiftDown   : Result := 'Shift down';
    tfRotateLeft  : Result := 'Rotate left';
    tfRotateRight : Result := 'Rotate right';
    tfMirror      : Result := 'Mirror';
    tfFlip        : Result := 'Flip';
    tfMove        : Result := 'Move selection';
    tfDrop        : Result := 'Drop selection';
    tfColor       : Result := 'Color change';
    tfClear       : Result := 'Cut/clear selection';
    tfPaste       : Result := 'Paste selection';
    tfSelect      : Result := 'Select';
    ELSE            Result := 'Drawing tool';
  end;
end;

function TUndoObj.GetUndoName: String;
begin
  IF fTransform = tfTool THEN
    begin
      IF fTool = '' THEN
        Result := 'Drawing' // should not happen
      ELSE Result := fTool;
    end
  ELSE Result := UndoNameOf(fTransform);
end;

{ TImageFrame }

procedure TImageFrame.FrameResize(Sender: TObject);
begin
  CalcCells;
end;

procedure TImageFrame.FrameMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
//If this event occurs, the mouse has moved out of the
// actual drawing area
begin
  IF (fMouseX <> -1) OR (fMouseY <> -1) THEN
    begin
      fMouseX := -1;
      fMouseY := -1;
      IF Assigned(fECellMove) THEN
        fECellMove(Self, fMouseX, fMouseY, 0);
    end;
end;

procedure TImageFrame.pbDrawMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
//Translate this mouse-down into a 'cell-down' event
VAR
  clipR  : TRect;
  Pt     : TPoint;
begin
  //Ignore mouse-down if other mouse button is already down
  CASE Button OF
    mbLeft  : IF fGotDown = 2  THEN Exit ELSE fGotDown := 1;
    mbRight : IF fGotDown = 1  THEN Exit ELSE fGotDown := 2;
    ELSE      Exit;
  end;
  //Restrict the cursor to the drawing area
  clipR := Bounds(Self.left + pbDraw.Left, Self.Top + pbDraw.Top,
    fCellAll - 4, fCellAll - 4);
  clipR.Topleft     := ClientToScreen(clipR.TopLeft);
  clipR.BottomRight := ClientToScreen(clipR.BottomRight);
  ClipCursor(@clipR);
  //Force the cursor into the restricted drawing area
  GetCursorPos(Pt);
  SetCursorPos(Pt.X, Pt.Y);
  IF CellFmPixel(X, Y, fMouseX, fMouseY) AND Assigned(fECellDown) THEN
    fECellDown(Self, fMouseX, fMouseY, Ord(Button));
end;

procedure TImageFrame.pbDrawMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
//Translate this mouse-move event into a 'cell-move' event, as
// long as the cell location IS actually changed
VAR cX, cY, Stat : Integer;
begin
  CellFmPixel(X, Y, cX, cY);
  //Stats is 1 if left button down, 2 if right, 0 if neither
  Stat := fGotDown;
  //Set 8-bit in Stat if Ctrl key is down
  IF ssCtrl IN Shift THEN Stat := Stat OR 8;
  //If cell-location has changed...
  IF (cX <> fMouseX) OR (cY <> fMouseY) THEN
    begin
      fMouseX := cX;
      fMouseY := cY;
      IF Assigned(fECellMove) THEN
        fECellMove(Self, fMouseX, fMouseY, Stat);
    end;
end;

procedure TImageFrame.pbDrawMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
//Translate the mouse-up event into a 'cell-up' event
VAR OtherButton : Boolean;
begin
  //Remove cursor restriction
  ClipCursor(nil);
  //fGotDown could possibly be 0 if middle button was down?
  IF fGotDown = 0 THEN Exit;
  CASE Button OF
    mbLeft  : OtherButton := fGotDown = 2;
    mbRight : OtherButton := fGotDown = 1;
    ELSE      OtherButton := True;
  END;
  CellFmPixel(X, Y, fMouseX, fMouseY);
  IF Assigned(fECellUp) THEN
    fECellUp(Self, fMouseX, fMouseY, Integer(OtherButton));
  fGotDown := 0;
end;

procedure TImageFrame.pbDrawPaint(Sender: TObject);
VAR co, ro : Integer;
begin
  WITH (Sender AS TPaintBox).Canvas DO
    begin
      SelectPalette(Handle, fPal.PalO, FALSE);
      RealizePalette(Handle);
      Pen.Color := clBlack;
      //Draw the grid
      FOR co := 0 TO fDimension DO
        begin
          MoveTo((co*fCellOne), 0);
          LineTo((co*fCellOne), fCellAll);
          MoveTo(0, (co*fCellOne));
          LineTo(fCellAll, (co*fCellOne));
        end;
      //Draw the cells
      FOR co := 0 TO fDimension-1 DO
        FOR ro := 0 TO fDimension-1 DO
          PaintOne(co, ro, -1, False);
      //Draw the focus rectangle
      IF NOT IsRectEmpty(fFoRect) THEN
        ShowFocusRect(True);
      //Draw the selection
      IF fMoving THEN
        WITH fFoRect DO
          FOR co := Left TO Right DO
            FOR ro := Top TO bottom DO
              JustPaint(co, ro, fCellsM[co-Left, ro-Top]);
    end;
end;

procedure TImageFrame.SetCellsM(const Value: TCellsType);
VAR co, ro : Integer;
begin
  fCellsM := Value;
  fMoving := True;
  FOR ro := fFoRect.Top TO fFoRect.Bottom DO
    FOR co := fFoRect.Left TO fFoRect.Right DO
      JustPaint(co, ro, fCellsM[co,ro]);
end;

function TImageFrame.GetColorCode: Integer;
begin
  CASE fColorCount OF
    2  : Result := 2;
    16 : Result := 1;
    ELSE Result := 0;
  end;
end;

procedure TImageFrame.SetColorCode(const Value: Integer);
begin
  CASE Value OF
    2 :  begin
           fColorCount := 2;
           fPalColors  := DefPal2;
         end;
    1 :  begin
           fColorCount := 16;
           fPalColors  := DefPal16;
         end;
    ELSE begin
           fColorCount := 256;
           fPalColors  := DefPal256;
         end;
  end;
end;

function TImageFrame.GetDimension: Integer;
begin
  IF NOT (fDimension IN [16, 32, 48]) THEN fDimension := 32;
  Result := fDimension;
end;

procedure TImageFrame.SetDimension(const Value: Integer);
// A Frame has no "OnCreate" event. In this frame, initialization
// of many fields takes place when the Dimension is set
begin
  fGotDown  := 0;
  fExists   := False;
  fMoving   := False;
  fUndoing  := False;
  fUndoList := TList.Create;
  ClearCells;
  CASE Value OF
    16, 32, 48 : fDimension := Value;
    ELSE Raise(Exception.Create('Invalid dimension'));
  end;
  fDMax         := fDimension - 1;
end;

procedure TImageFrame.SetFocusRect(const Value: TRect);
begin
  IF fMoving THEN FocusCellsDrop;
  IF NOT IsRectEmpty(fFoRect) THEN ShowFocusRect(False);
  fFoRect := Value;
  IF NOT IsRectEmpty(fFoRect) THEN ShowFocusRect(True);
  fMoving := False;
end;

function TImageFrame.GetImageBlank: Boolean;
//Returns True only if image contains nothing but transparent pixels
VAR co, ro : Integer;
begin
  Result := False;
  FOR co := 0 TO fDimension-1 DO
    FOR ro := 0 TO fDimension-1 DO
      IF fCells[co,ro] <> palTrans THEN
        Exit;
  Result := True;
end;

function TImageFrame.GetPixelCol(X, Y: Integer): SmallInt;
begin
  IF fMoving AND PtInCoRect(fFoRect, X, Y) THEN
    Result := fCellsM[X-fFoRect.left, Y-fFoRect.Top]
  ELSE Result := fCells[X,Y];
end;

function TImageFrame.GetUndoName: String;
// Returns a name for the top item in the Undo stack
begin
  Result := '&Undo';
  IF fUndoList.Count = 0 THEN Exit;
  Result := Result + ' ' +
    TUndoObj(fUndoList[fUndoList.count-1]).UndoName;
end;

procedure TImageFrame.CalcCells;
VAR D, Margin : Integer;
begin
  IF csDesigning IN ComponentState THEN Exit;
  //Set D to the smaller of Width and Height, then subtract 4
  D := Width;
  IF Height < D THEN D := Height;
  Dec(D, 4);
  //Set fCellOne
  fCellOne := D DIV GetDimension;
  IF (fCellOne * fDimension) + 8 > D THEN Dec(fCellOne);
  IF fCellOne < 4 THEN
    begin
      ShowMessage('Error: cell size less than 4 pixels. '+
        'Resetting to 4.');
      fCellOne := 4;
    end;
  //Set fAdj - used in drawing XOR-rects and XOR-ellipses
  IF Odd(fCellOne) THEN
    fAdj        := 2
  ELSE fAdj     := 1;
  fCellHalf     := fCellOne DIV 2;
  fCellAll      := (fCellOne*fDimension)+1;
  Margin        := (D - fCellAll) DIV 2;
  pbDraw.Left   := Margin+2;
  pbDraw.Top    := Margin+2;
  pbDraw.Width  := fCellAll;
  pbDraw.Height := fCellAll;
end;

function TImageFrame.CellFmPixel(X, Y: Integer; VAR cX, cY: Integer):
  Boolean;
// Take an X,Y pixel coordinate and put the corresponding
// cell within the drawing area into cX,cY. Return false if
// outside the drawing grid.
begin
  IF (X < 0) OR (Y < 0) THEN
    begin cX := -1; cY := -1; end
  ELSE IF (X >= fCellAll) OR (Y >= fCellAll) THEN
    begin cX := -1; cY := -1; end
  ELSE
    begin
      cX := X DIV fCellOne;
      cY := Y DIV fCellOne;
    end;
  Result := ValidCell(cX, cY);
  IF NOT Result THEN
    begin cX := -1; cY := -1; end;
end;

procedure TImageFrame.ShowFocusRect(Show : Boolean);
// If Show is true, draw a focus rectangle; if not, draw a plain
// black rectangle. If the selection extends beyond the edge
// of the drawing area, don't draw that edge in the focus rect
VAR
  ixL, ixT, ixR, ixB : Integer;
  unL, unT, unR, unB : Boolean;
begin
  WITH fFoRect DO
    begin
      ixL := Left;   unL := ixL < 0;     IF unL THEN ixL := 0;
      ixR := Right;  unR := ixR > fDMax; IF unR THEN ixR := fDMax;
      ixT := Top;    unT := ixT < 0;     IF unT THEN ixT := 0;
      ixB := Bottom; unB := ixB > fDMax; IF unB THEN ixB := fDMax;
    end;
  ixL := ixL*fCellOne;
  ixR := Succ(ixR)*fCellOne;
  ixT := ixT *fCellOne;
  ixB := Succ(ixB)*fCellOne;
  WITH pbDraw.Canvas DO
    begin
      Brush.Style := bsClear;
      Pen.Style   := psSolid;
      IF Show THEN
        begin
          Pen.Color := clWhite;
          MoveTo(ixL, ixT);
          IF unL THEN MoveTo(ixL, ixB) ELSE LineTo(ixL, ixB);
          IF unB THEN MoveTo(ixR, ixB) ELSE LineTo(ixR, ixB);
          IF unR THEN MoveTo(ixR, ixT) ELSE LineTo(ixR, ixT);
          IF unT THEN MoveTo(ixL, ixT) ELSE LineTo(ixL, ixT);
          Pen.Color := clBlack;
          Pen.Style := psDot;
          MoveTo(ixL, ixT);
          IF unL THEN MoveTo(ixL, ixB) ELSE LineTo(ixL, ixB);
          IF unB THEN MoveTo(ixR, ixB) ELSE LineTo(ixR, ixB);
          IF unR THEN MoveTo(ixR, ixT) ELSE LineTo(ixR, ixT);
          IF unT THEN MoveTo(ixL, ixT) ELSE LineTo(ixL, ixT);
          Pen.Style := psSolid;
        end
      ELSE
        begin
          Pen.Color := clBlack;
          MoveTo(ixL, ixT);
          IF unL THEN MoveTo(ixL, ixB) ELSE LineTo(ixL, ixB);
          IF unB THEN MoveTo(ixR, ixB) ELSE LineTo(ixR, ixB);
          IF unR THEN MoveTo(ixR, ixT) ELSE LineTo(ixR, ixT);
          IF unT THEN MoveTo(ixL, ixT) ELSE LineTo(ixL, ixT);
        end;
  end;
end;

function TImageFrame.ValidCell(vCo, vRo: Integer): Boolean;
begin
  Result := (vCo >= 0) AND (vRo >= 0) AND
    (vCo < fDimension) AND (vRo < fDimension);
end;

procedure TImageFrame.ActChangeCol;
//replace foreground with background
VAR co, ro : Integer;
  TUO  : TUndoObj;
begin
  TUO := UndoWRect(fFoRect, tfColor);
  IF IsRectEmpty(fFoRect) THEN
    begin
      FOR co := 0 TO fDMax DO
        FOR ro := 0 TO fDMax DO
          IF fCells[co,ro] = fPal.fore THEN
            PaintOne(co, ro, fPal.Back, True);
    end
  ELSE
    begin
      WITH fFoRect DO
        FOR co := Left TO Right DO
          FOR ro := Top TO Bottom DO
            IF fCellsM[co-Left, ro-Top] = fPal.Fore THEN
              begin
                TUO.AddCell(co, ro, fPal.Fore);
                fCellsM[co-Left, ro-Top] :=
                  fPal.Back;
                JustPaint(co, ro, fPal.Back);
              end;
    end;
end;

procedure TImageFrame.ActShift(typ : Integer);
// "un" is true if we're performing an undo (and hence should not
//   store additional undo information
  procedure ShiftSelf;
  VAR
    co, ro : Integer;
    TUO    : TUndoObj;
  begin
    IF NOT fUnDoing THEN // Save UNDO information
      CASE typ OF
        0 : begin //left - save left column
          TUO := UndoTrans(tfShiftLeft);
          FOR ro := 0 TO Dimension-1 DO
            TUO.AddCell(0, ro, fCells[0,ro]);
        end;
        1 : begin //right - save right column
          TUO := UndoTrans(tfShiftRight);
          FOR ro := 0 TO Dimension-1 DO
            TUO.AddCell(Dimension-1, ro, fCells[Dimension-1,ro]);
        end;
        2 : begin //up - save top row
          TUO := UndoTrans(tfShiftUp);
          FOR co := 0 TO Dimension-1 DO
            TUO.AddCell(co, 0, fCells[co, 0]);
        end;
        3 : begin //down - save bottom row
          TUO := UndoTrans(tfShiftDown);
          FOR co := 0 TO Dimension-1 DO
            TUO.AddCell(co, Dimension-1, fCells[co, Dimension-1]);
        end;
      END;
    CASE typ OF
      0 : FOR ro := 0 TO Dimension-1 DO //left
            FOR co := 1 TO Dimension-1 DO
              fCells[co-1, ro] := fCells[co, ro];
      1 : FOR ro := 0 TO Dimension-1 DO //right
            FOR co := Dimension-2 DOWNTO 0 DO
              fCells[co+1, ro] := fCells[co, ro];
      2 : FOR co := 0 TO Dimension-1 DO //up
            FOR ro := 1 TO Dimension-1 DO
              fCells[co, ro-1] := fCells[co, ro];
      3 : FOR co := 0 TO Dimension-1 DO //down
            FOR ro := Dimension-2 DOWNTO 0 DO
              fCells[co, ro+1] := fCells[co, ro];
    END;
    IF NOT fUnDoing THEN // Clear newly-exposed row or column
      CASE typ OF
        0 : FOR ro := 0 TO Dimension-1 DO //left
              fCells[Dimension-1, ro] := fPal.back;
        1 : FOR ro := 0 TO Dimension-1 DO //right
              fCells[0, ro] := fPal.back;
        2 : FOR co := 0 TO Dimension-1 DO //up
              fCells[co, Dimension-1] := fPal.back;
        3 : FOR co := 0 TO Dimension-1 DO //down
              fCells[co, 0] := fPal.back;
      END;
    pbDraw.Refresh;
  end;

  procedure ShiftRect;
  VAR dX, dY : Integer;
  begin
    UndoWRect(fFoRect, tfShift);
    FocusCellsInit;
    dX := 0;
    dY := 0;
    CASE typ OF
      0 : begin dX := -1; dY :=  0; end;
      1 : begin dX :=  1; dY :=  0; end;
      2 : begin dX :=  0; dY := -1; end;
      3 : begin dX :=  0; dY :=  1; end;
    END;
    FocusCellsMove(fFoRect.Left+dX, fFoRect.Top+dY);
  end;

begin
  IF IsRectEmpty(fFoRect) THEN
    ShiftSelf
  ELSE ShiftRect;
end;

procedure TImageFrame.ActTransform(typ: Integer);
// "un" is true if we're performing an undo (and hence should not
//   store additional undo information
  procedure TransSelf;
  VAR co, ro : Integer;
  begin
    CASE typ OF
      0 : FOR ro := 0 TO Dimension-1 DO //rotate left
            FOR co := 0 TO Dimension-1 DO
              fCellsM[ro, Dimension-1-co] := fCells[co, ro];
      1 : FOR co := 0 TO Dimension-1 DO //rotate right
            FOR ro := 0 TO Dimension-1 DO
              fCellsM[Dimension-1-ro, co] := fCells[co,ro];
      2 : FOR co := 0 TO Dimension-1 DO //mirror
            FOR ro := 0 TO Dimension-1 DO
              fCellsM[Dimension-1-co, ro] := fCells[co,ro];
      3 : FOR co := 0 TO Dimension-1 DO //flip
            FOR ro := 0 TO Dimension-1 DO
              fCellsM[co, Dimension-1-ro] := fCells[co,ro];
    END;
    fCells := fCellsM;
    pbDraw.Refresh;
  end;

  procedure TransRect;
  VAR
    co, ro  : Integer;
    NuRect  : TRect;
    NuCells : TCellsType;
    function  RotateRect(R : TRect) : TRect;
    begin
      WITH R DO
        begin
          // Result is R rotated about its middle
          Result.Left   := ((Right+Left)DIV 2)-((Bottom-Top) DIV 2);
          Result.Top    := ((Bottom+Top)DIV 2)-((Right-Left) DIV 2);
          Result.Right  := Result.Left + Bottom - Top;
          Result.Bottom := Result.Top + Right - Left;
        end;
      WITH Result DO
        begin
          IF Left > fDMax THEN  OffsetRect(Result, fDMax-Left, 0);
          IF Right < 0 THEN  OffsetRect(Result, 0-Right, 0);
          IF Top > fDMax THEN   OffsetRect(Result, 0, fDMax-Top);
          IF Bottom < 0 THEN OffsetRect(Result, 0, 0-Bottom);
        end;
    end;

  begin
//    FocusCellsInit; 0.11
    IF typ < 2 THEN NuRect := RotateRect(fFoRect)
    ELSE NuRect := fFoRect;
    WITH fFoRect DO
      CASE typ OF
        0 : FOR ro := 0 TO Bottom-Top DO //rotate left
              FOR co := 0 TO Right-Left DO
                NuCells[ro, Right-Left-co] := fCellsM[co, ro];
        1 : FOR ro := 0 TO Bottom-Top DO //rotate right
              FOR co := 0 TO Right-Left DO
                NuCells[Bottom-Top-ro, co] := fCellsM[co,ro];
        2 : FOR ro := 0 TO Bottom-Top DO //mirror
              FOR co := 0 TO Right-Left DO
                NuCells[Right-Left-co, ro] := fCellsM[co,ro];
        3 : FOR ro := 0 TO Bottom-Top DO //flip
              FOR co := 0 TO Right-Left DO
                NuCells[co, Bottom-Top-ro] := fCellsM[co,ro];
      END;
    ShowFocusRect(False);
    FOR co := fFoRect.Left TO fFoRect.Right DO
      FOR ro := fFoRect.Top TO fFoRect.Bottom DO
        PaintOne(co, ro, -1, False); 
    fCellsM := NuCells;
    fFoRect := NuRect;
    ShowFocusRect(True);
    FocusCellsMove(fFoRect.Left, fFoRect.Top);
  end;

begin
  IF NOT fUnDoing THEN
    CASE Typ OF
      0 : UndoTrans(tfRotateLeft );
      1 : UndoTrans(tfRotateRight);
      2 : UndoTrans(tfMirror     );
      3 : UndoTrans(tfFlip       );
    END;
  IF IsRectEmpty(fFoRect) THEN
    TransSelf
  ELSE TransRect;
end;

function TImageFrame.CenterOf(I : Integer) : Integer;
begin
  Result := (I*fCellOne) + fCellHalf;
end;

procedure TImageFrame.ClearCells;
VAR N, ro, co : Integer;
begin
  FOR N := fUndoList.Count-1 DOWNTO 0 DO
    TUndoObj(fUndoList[N]).Free;
  fUndoList.Clear;
  fFoRect := NoRect;
  FOR ro := 0 TO 47 DO
    FOR co := 0 TO 47 DO
      fCells[co,ro] := palTrans;
  FocusRect := NoRect;
  fExists := False;
end;

procedure TImageFrame.ColorToPal(co, ro: Integer; isFore: Boolean);
begin
  IF IsFore THEN
    fPal.Fore := fCells[co, ro]
  ELSE
    fPal.Back := fCells[co, ro];
end;

function TImageFrame.ColString(cX,cY : Integer): String;
begin
  IF NOT ValidCell(cX,cY) THEN Result := ''
  ELSE IF PtInCoRect(fFoRect, cX, cY) THEN
    Result := StringForColor(fPal.Pal,
      fCellsM[cX-fFoRect.Left, cY-fFoRect.Top])
    ELSE Result := StringForColor(fPal.Pal,
      fCells[cX,cY]);
end;

procedure TImageFrame.FocusCellsDrop;
VAR co, ro : Integer;
begin
  IF NOT fMoving THEN Exit;
  IF NOT IsRectEmpty(fFoRect) THEN
    begin
      IF NOT fUndoing THEN UndoWRect(fFoRect, tfDrop);
      WITH fFoRect DO
        FOR co := Left TO Right DO
          FOR ro := Top TO Bottom DO
            PaintOne(co, ro, fCellsM[co-fFoRect.Left,
              ro-fFoRect.Top], NOT fUndoing);
    end;
  fMoving := False;
end;

procedure TImageFrame.FocusCellsPreClear;
VAR co, ro : Integer;
  TUO : TUndoObj;
begin
  IF IsRectEmpty(fFoRect) THEN Exit;
  IF fMoving THEN Exit;
  TUO := UndoWRect(fFoRect, tfClear);
  WITH fFoRect DO
    FOR co := Left TO Right DO
      FOR ro := Top TO Bottom DO
        TUO.AddCell(co, ro, fCells[co,ro]);
end;

procedure TImageFrame.FocusCellsClear;
VAR co, ro : Integer;
  TUO : TUndoObj;
begin
  IF IsRectEmpty(fFoRect) THEN Exit;
  IF NOT fMoving THEN Exit;
  TUO := UndoWRect(fFoRect, tfClear);
  WITH fFoRect DO
    FOR co := Left TO Right DO
      FOR ro := Top TO Bottom DO
        begin
          TUO.AddCell(co, ro, fCellsM[co-Left, ro-Top]);
          PaintOne(co, ro, -1, False);
        end;
  ShowFocusRect(False);
  fFoRect := NoRect;
end;

procedure TImageFrame.FocusCellsGet;
VAR co, ro : Integer;
begin
  IF fMoving THEN Exit;
  WITH fFoRect DO
    FOR co := Left TO Right DO
      FOR ro := Top TO Bottom DO
        fCellsM[co-fFoRect.Left, ro-fFoRect.Top] :=
          fCells[co, ro];
end;

procedure TImageFrame.FocusCellsInit;
VAR co, ro : Integer;
begin
  IF NOT fMoving THEN
    begin
      WITH fFoRect DO
        FOR co := Left TO Right DO
          FOR ro := Top TO Bottom DO
            begin
              fCellsM[co-fFoRect.Left, ro-fFoRect.Top] :=
                fCells[co, ro];
              PaintOne(co, ro, fPal.Back, {0.11True} False);
            end;
    end;
  fMoving := True;
  FocusCellsMove(fFoRect.Left, fFoRect.Top);
end;

procedure TImageFrame.FocusCellsMove(dX, dY: Integer);
VAR
  co, ro : Integer;
  NuRect : TRect;
begin
  NuRect := fFoRect;
  OffsetRect(NuRect, dX-NuRect.Left, dY-NuRect.Top);
  ShowFocusRect(False);
  WITH fFoRect DO
    FOR co := Left TO Right DO
      FOR ro := Top TO bottom DO
        IF PtInCoRect(fFoRect, co, ro) AND
          (NOT PtInCoRect(NuRect, co, ro)) THEN
            PaintOne(co, ro, -1, False);
  fFoRect := NuRect;
  WITH fFoRect DO
    FOR co := Left TO Right DO
      FOR ro := Top TO bottom DO
        JustPaint(co, ro, fCellsM[co-Left, ro-Top]);
  ShowFocusRect(True);
end;

procedure TImageFrame.JustPaint(co, ro, colr : Integer);
// Paint a cell on-screen without changing its stored color
begin
  IF NOT ValidCell(co, ro) THEN Exit;
  WITH pbDraw.Canvas DO
    begin
      SelectPalette(Handle, fPal.PalO, FALSE);
      RealizePalette(Handle);
      CASE Colr OF
        0..palMaxim : begin
          Brush.Color := fPal.Pal[Colr];
          FillRect(Bounds((co*fCellOne)+1,
            (ro*fCellOne)+1, fCellOne-1, fCellOne-1));
        end;
        palTrans : begin
          Brush.Color := clBtnFace;
          FillRect(Bounds((co*fCellOne)+1,
            (ro*fCellOne)+1, fCellOne-1, fCellOne-1));
          Pen.Color := clBlue;
          MoveTo((co*fCellOne)+2, (ro*fCellOne)+2);
          LineTo(((co+1)*fCellOne)-1, ((ro+1)*fCellOne)-1);
        end;
        palInvrt : begin
          Brush.Color := clBtnFace;
          FillRect(Bounds((co*fCellOne)+1,
            (ro*fCellOne)+1, fCellOne-1, fCellOne-1));
          Pen.Color := clRed;
          MoveTo((co*fCellOne)+2,     ((ro+1)*fCellOne)-1);
          LineTo(((co+1)*fCellOne)-1, (ro*fCellOne)+2);
        end;
      end;
    end;
end;

procedure TImageFrame.PaintOne(co, ro, colr: Integer; un : Boolean);
// "un" is true if we should store undo information
VAR WasColr : Integer;
begin
  IF NOT ValidCell(co, ro) THEN Exit;
  WasColr := fCells[co,ro];
  IF colr >= 0 THEN fCells[co,ro] := colr
  ELSE colr := fCells[co,ro];
  JustPaint(co, ro, colr);
  IF Un AND (NOT fUndoing) AND (wasColr <> Colr) AND (fUndoList.Count > 0) THEN
    begin
      IF fUndoList.Count = 0 THEN
        WasColr := WasColr;
    WITH TUndoObj(fUndoList[fUndoList.Count-1]) DO
      begin
        AddCell(co, ro, WasColr);
      end;
    end;
end;

function TImageFrame.PtInCoRect(R : TRect; X, Y: Integer): Boolean;
begin
  IF IsRectEmpty(R) THEN Result := False
  ELSE WITH R DO
    Result := PtInRect(Rect(Left, Top, Right+1, Bottom+1),
      Point(X, Y));
end;

procedure TImageFrame.Setup(D, M, U: TCellEvent; P: TPalFrame2);
begin
  fECellDown := D;
  fECellMove := M;
  fECellUp   := U;
  fPal       := P;
end;

function TImageFrame.UndoAble: Boolean;
begin
  Result := fUndoList.Count > 0;
end;

procedure TImageFrame.UndoDoIt;
VAR
  TUO          : TUndoObj;
  co, ro, colr : Integer;
begin
  fUndoing := True;
  try
    TUO := TUndoOBj(fUndoList[fUndoList.Count-1]);
    fUndoList.Delete(fUndoList.Count-1);
    CASE TUO.fTransform OF
      tfShiftLeft  : ActShift(1);
      tfShiftRight : ActShift(0);
      tfShiftUp    : ActShift(3);
      tfShiftDown  : ActShift(2);
      tfRotateLeft   : ActTransform(1);
      tfRotateRight  : ActTransform(0);
      tfMirror       : ActTransform(2);
      tfFlip         : ActTransform(3);
      tfSelect       : begin
        FocusRect := TUO.FocusRect;
      end;
      tfMove         : begin
        FocusCellsMove(TUO.FocusRect.Left,
          TUO.FocusRect.Top);
      end;
      tfDrop         : begin
        fFoRect := TUO.FocusRect;
        fMoving := False;
        FocusCellsGet;
        fMoving := True;
        REPEAT
          TUO.RemCell(co, ro, colr);
          IF co >= 0 THEN
            PaintOne(co, ro, Colr, False);
        UNTIL co < 0;
        pbDraw.Refresh;
      end;
      tfClear : begin
        fFoRect := TUO.FocusRect;
        REPEAT
          TUO.RemCell(co, ro, colr);
          IF co >= 0 THEN
            fCellsM[co-fFoRect.Left, ro-fFoRect.Top] := Colr;
        UNTIL co < 0;
        pbDraw.Refresh;
      end;
      tfPaste : begin
        IF NOT IsRectEmpty(fFoRect) THEN
          ShowFocusRect(False);
        WITH fFoRect DO
          FOR co := Left TO Right DO
            FOR ro := Top TO Bottom DO
              PaintOne(co, ro, -1, False);
        fFoRect := NoRect;
      end;
      tfColor : begin 
        IF IsRectEmpty(TUO.FocusRect) THEN
          REPEAT
            TUO.RemCell(co, ro, colr);
            IF co >= 0 THEN
              PaintOne(co, ro, Colr, False);
          UNTIL co < 0
        ELSE
          WITH TUO.FocusRect DO
            REPEAT
              TUO.RemCell(co, ro, colr);
              IF co >= 0 THEN
                begin
                  fCellsM[co-Left, ro-Top] := colr;
                  JustPaint(co, ro, Colr);
                end;
            UNTIL co < 0 ;
      end;
    END;
    CASE TUO.fTransform OF
      tfShiftLeft,
      tfShiftRight,
      tfShiftUp,
      tfShiftDown,
      tfTool : begin
        REPEAT
          TUO.RemCell(co, ro, colr);
          IF co >= 0 THEN
            PaintOne(co, ro, Colr, False);
        UNTIL co < 0;
      end;
      ELSE IF TUO.fList.Count > 0 THEN
        Raise(Exception.Create('Blargh!'));
    END;
    TUO.Free;
  finally
    fUndoing := False;
  end;
end;

procedure TImageFrame.UndoMaxReached;
begin
  IF fUndoList.Count >= UNDOMAX THEN
    begin
      TUndoObj(fUndoList[0]).Free;
      fUndoList.Delete(0);
    end;
end;

function TImageFrame.UndoTrans(vTransform : TTransform) : TUndoObj;
begin
  OutputDebugString(PChar(Format('UndoTrans: %s',
    [TUndoObj.UndonameOf(vTransform)])));
  UndoMaxReached;
  Result := TUndoObj.Create(vTransform, '');
  fUndoList.Add(Result);
end;

function TImageFrame.UndoTool(const vTool : String) : TUndoObj;
VAR TUO : TUndoObj;
begin
  OutputDebugString(PChar(Format('UndoTool: %s',
    [vTool])));
  IF fUndoList.Count > 0 THEN
    begin
      TUO := tUndoObj(fUndoList[fUndoList.Count-1]);
      IF TUO.fTransform = tfSelect THEN
        begin
          OutputDebugString('(note: removed tfSelect)');
          fUndoList.Delete(fUndoList.Count-1);
          TUO.Free;
        end;
    end;
  UndoMaxReached;
  Result := TUndoObj.Create(tfTool, vTool);
  fUndoList.Add(Result);
end;

function TImageFrame.UndoWRect(R : TRect; vTransform : TTransform) : TUndoObj;
VAR TUO : TUndoObj;
begin
  OutputDebugString(PChar(Format('UndoWRect: starting %s',
    [TUndoobj.UndoNameOf(vTransform)])));
  Result := nil;
  IF fUndoList.Count = 0 THEN TUO := nil
  ELSE TUO := tUndoObj(fUndoList[fUndoList.Count-1]);
  // Don't record a drop immediately after a select
  IF (vTransform = tfDrop) AND (TUO <> nil) AND
    (TUO.fTransform = tfSelect) THEN Exit;
  // Convert multiple shifts into a single move
  IF (vTransform = tfShift) THEN
    begin
      IF TUO = nil THEN Exit
      ELSE IF TUO.fTransform <> tfMove THEN
        vTransform := tfMove
      ELSE Exit;
    end;
  OutputDebugString(PChar(Format('UndoWRect: %s (%d,%d,%d,%d)',
    [TUndoobj.UndoNameOf(vTransform),
    R.Left, R.Top, R.Right, R.Bottom])));
  UndoMaxReached;
  Result := TUndoObj.Create(vTransform, '');
  Result.FocusRect := R;
  fUndoList.Add(Result);
  OutputDebugString('UndoWRect completed');
end;

end.
