unit palfrm2u;

// Copyright  2000 by Ziff-Davis, Inc.
// Written by Neil J. Rubenking

// This Delphi 5 TFrame descendant encapsulates the palette
// color-picker, thereby keeping its complexity out of the
// main form
// This is a modified version of the palette control originally
// devised for CursorDance. This version adds support for
// 2-color palette, and lines up the 256-color palette in
// six rows

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, palconsu, StdCtrls;

type
  TPalFrame2 = class(TFrame)
    pnlPal : TPanel;
    pbPal  : TPaintBox;
    procedure pbPalMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure pbPalMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure pbPalPaint(Sender: TObject);
  private
    { Private declarations }
    PalM              : Integer; // margin for palette display
    thePalColor       : Integer; // under-cursor palette color
    fForeColor        : Integer; // foreground color
    fBackColor        : Integer; // background color
    palR              : Integer; // # rows in palette display
    palC              : Integer; // # cols in palette display
    palS              : Integer; // cell size in palette display
    palW              : Integer; // pixel width for palette display
    palH              : Integer; // pixel height for palette display
    palX              : Integer; // start-offset for palette
    fPalCode          : Integer; // 0=256, 1=16, 2=2 colors
    aPalette          : TPalArray; // array of palette entries
    PaletteObj        : HPalette;  // palette handle from aPalette
    procedure SetAPalette(value : TPalArray);
    procedure SetBackColor(const Value: Integer);
    procedure SetForeColor(const Value: Integer);
  public
    { Public declarations }
    OnColorChange : TNotifyEvent;
    procedure SetPalValues(PalCode : Integer);
    property Pal  : TPalArray read aPalette write SetAPalette;
    property PalO : HPalette read PaletteObj;
    property fore : Integer read fForeColor write SetForeColor;
    property back : Integer read fBackColor write SetBackColor;
  end;

implementation

{$R *.DFM}
const
  palMaxim = 255;
  palTrans = 256;
  palInvrt = 257;

procedure TPalFrame2.pbPalMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
VAR Xcell, Ycell : Integer;
begin
  thePalColor               := -1; // default - no color
  pbPal.Hint := '';                // default - no status
  Xcell                        := (X-palM) DIV palS;
  Ycell                        := (Y-palM) DIV palS;
  IF X < PalM      THEN Exit; // outside, to left
  IF Y < PalM      THEN Exit; // outside, to top
  IF Ycell >= palR THEN Exit; // outside, to bottom
  IF Xcell > palC  THEN Exit; // outside, to right
  IF Xcell = palC  THEN       // in last "special" column
    begin
      IF Ycell = 0 THEN
        begin
          pbPal.Hint  := 'transparent';
          thePalColor := palTrans;
        end
      ELSE IF Ycell = palR-1 THEN
        begin
          pbPal.Hint  := 'inverted';
          thePalColor := palInvrt;
         end;
      Exit;
    end;
  // In 256-color mode, there are cell-locations beyond the
  // last in-use cell...
  IF Ycell*palC+Xcell-PalX > palMaxim THEN Exit;
  //(XCell, YCell) is within the palette rectangle
  thePalColor := Ycell*palC+Xcell-PalX;
  pbPal.Hint := StringForColor(aPalette, thePalColor);
end;

procedure TPalFrame2.pbPalMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  IF thePalColor < 0 THEN Exit;
  CASE Button OF
    mbLeft  : IF thePalColor <> fForeColor THEN
      begin
        fForeColor := thePalColor;
        IF Assigned(OnColorChange) THEN OnColorChange(Self);
      end;
    mbRight : IF thePalColor <> fBackColor THEN
      begin
        fBackColor := thePalColor;
        IF Assigned(OnColorChange) THEN OnColorChange(Self);
      end;
  end;
end;

procedure TPalFrame2.pbPalPaint(Sender: TObject);
VAR co, ro, M1 : Integer;
begin
  M1 := PalM-1;
  WITH pbPal.Canvas DO
    begin
      SelectPalette(Handle, PaletteObj, FALSE);
      RealizePalette(Handle);
      Pen.Color := clBtnShadow;
      // Box the transparent and inverted cells
      Rectangle(M1+PalC*PalS, M1, M1+(PalC+1)*PalS+1, M1+PalS+1);
      Rectangle(M1+PalC*PalS, M1+PalH-PalS-1, M1+(PalC+1)*PalS+1,
        M1+PalH);
      // Fill in the palette colors
      FOR ro := 0 TO palR-1 DO
        FOR co := 0 TO palC-1 DO
          begin
            IF (ro*palC)+co < palX THEN Continue;
            IF (ro*palC)+co-palX > 255 THEN Continue;
            Brush.Color := aPalette[(ro*palC)+co-palX];
            FillRect(Bounds(PalM+(co*palS),
              PalM+(ro*palS), palS-1, palS-1));
            end;
      // Draw blue backslash for transparent
      Pen.Color := clBlue;
      MoveTo(PalM+palC*palS+1, PalM+1);
      LineTo(PalM+(palC+1)*palS-2, PalM+palS-2);
      // Draw red slash for inverted
      Pen.Color := clRed;
      MoveTo(PalM+palC*palS+1, PalM+palH-4);
      LineTo(PalM+(palC+1)*palS-2, PalM+palH-palS-1);
    end;
end;

procedure TPalFrame2.SetAPalette(value: TPalArray);
VAR N : Integer;

  function GetPalFrom(P : TPalArray; colors : Integer) : HPalette;
  VAR
    PLP : PLogPalette;
    N   : Integer;
  begin
    GetMem(PLP, SizeOf(PLP)+Colors*SizeOf(TPaletteEntry));
    try
      PLP^.palVersion := $300;
      PLP^.palNumEntries := colors;
      {$IFOPT R+}
      {$R-}
      {$DEFINE WasPlus}
      {$ENDIF}
      FOR N := 0 TO colors-1 DO
        WITH PLP^.palPalEntry[N] DO
          begin
            peRed    := GetRValue(P[N]);
            peGreen  := GetGValue(P[N]);
            peBlue   := GetBValue(P[N]);
            peFlags  := 0;//PC_NOCOLLAPSE;
          end;
      {$IFDEF WasPlus}
      {$R+}
      {$ENDIF}
      Result := CreatePalette(PLP^);
    finally
      FreeMem(PLP);
    end;
  end;
begin
  aPalette := Value;
  FOR N := 0 TO 255 DO // Helps if forced into 256-color
    aPalette[N] := aPalette[N]+$02000000;
  fForeColor := 0;
  CASE fPalCode OF
    2 : fBackColor := 1;
    1 : fBackColor := 15;
    ELSE fBackColor := 255;
  END;
  PaletteObj := GetPalFrom(aPalette, fBackColor+1);
  // If first palette color is not black, seek black
  IF aPalette[fForeColor] <> $000000 THEN
    FOR N := 0 TO fBackColor DO
      IF aPalette[N] = $000000 THEN
        begin
          fForeColor := N;
          Break;
        end;
  // If last palette color is not white, seek white
  IF aPalette[fBackColor] <> $FFFFFF THEN
    FOR N := fBackColor DOWNTO 0 DO
      IF aPalette[N] = $FFFFFF THEN
        begin
          fBackColor := N;
          Break;
        end;
  pbPal.Refresh;
  IF Assigned(OnColorChange) THEN OnColorChange(Self);
end;

procedure TPalFrame2.SetBackColor(const Value: Integer);
begin
  IF fBackColor <> Value THEN
    begin
      fBackColor := Value;
      IF Assigned(OnColorChange) THEN OnColorChange(Self);
    end;
end;

procedure TPalFrame2.SetForeColor(const Value: Integer);
begin
  IF fForeColor <> Value THEN
    begin
      fForeColor := Value;
      IF Assigned(OnColorChange) THEN OnColorChange(Self);
    end;
end;

procedure TPalFrame2.SetPalValues(PalCode : Integer);
begin
  PalM       := 1;
  fPalCode   := PalCode;
  fForeColor := 0;
  CASE PalCode OF
    2 : begin
      palR       := 2;
      palC       := 1;
      palS       := 27;
      palX       := 0;
      fbackColor := 1;
    end;
    1 : begin
      palR       := 2;
      palC       := 8;
      palS       := 27;
      palX       := 0;
      fbackColor := 15;
    end;
    ELSE begin
      palR       := 6;
      palC       := 48;
      palS       := 9;
      palX       := 2;
      fbackColor := 255;
    end;
  END;
  palW := (palS*palC)+1;
  palH := (palS*palR)+1;
  IF Height < PalH + 6 THEN Height := PalH + 6;
  pbPal.Refresh;
  IF Assigned(OnColorChange) THEN OnColorChange(Self);
end;

end.
